/********************************************************************************
*                                                                               *
*                        P r i n t   J o b   D i a l o g                        *
*                                                                               *
*********************************************************************************
* Copyright (C) 1999 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXPrintDialog.cpp,v 1.7 1999/10/14 05:55:05 jeroen Exp $                 *
********************************************************************************/
#include "xincs.h"
#include "fx.h"


/*
  Notes:
  - Inputs print job:
    Destination Printer, or file (browse to file)
    Orientation
    Paper size [A4(210x297mm) B5, Letter(8.5x11in), Legal, Executive, ?...
    Print all pages, or range of pages [first/last]
    Page collate order [first page first, last page first, even/odd]
    Number of copies
    Color/gray/blackwhite
    Custom paper dimensions
    Standard paper dimensions

                  Page size          Print area         Top/Bottom   Sides
 
    Letter        8.5 x 11 in        8.2 x 10.6 in      0.25 in       0.25 in
    A4            210 x 297 mm       200 x 287 mm       5 mm         5 mm
    Legal         8.5 x 14 in        8.2 x 13.6 in      0.2 in       0.15 in
    #10 Envelope  4.13 x 9.5 in      3.8 x 9.2 in       0.16 in      0.14 in
                  104.8 x 241.3 mm   96.5 x 233.7 mm    4.1 mm       3.6 mm
                  
    Known paper sizes:
  
       PAPERSIZE    X"         Y"         X cm      Y cm
       -----------------------------------------------------
       11x17        11"        17"        27.94     43.18
       a0           33.0556"   46.7778"   83.9611   118.816
       a10          1.02778"   1.45833"   2.61056   3.70417
       a1           23.3889"   33.0556"   59.4078   83.9611
       a2           16.5278"   23.3889"   41.9806   59.4078
       a3           11.6944"   16.5278"   29.7039   41.9806
       a4           8.26389"   11.6944"   20.9903   29.7039
       a5           5.84722"   8.26389"   14.8519   20.9903
       a6           4.125"     5.84722"   10.4775   14.8519
       a7           2.91667"   4.125"     7.40833   10.4775
       a8           2.05556"   2.91667"   5.22111   7.40833
       a9           1.45833"   2.05556"   3.70417   5.22111
       archA        9"         12"        22.86     30.48
       archB        12"        18"        30.48     45.72
       archC        18"        24"        45.72     60.96
       archD        24"        36"        60.96     91.44
       archE        36"        48"        91.44     121.92
       b0           39.3889"   55.6667"   100.048   141.393
       b1           27.8333"   39.3889"   70.6967   100.048
       b2           19.6944"   27.8333"   50.0239   70.6967
       b3           13.9167"   19.6944"   35.3483   50.0239
       b4           9.84722"   13.9167"   25.0119   35.3483
       b5           6.95833"   9.84722"   17.6742   25.0119
       flsa         8.5"       13"        21.59     33.02
       flse         8.5"       13"        21.59     33.02
       halfletter   5.5"       8.5"       13.97     21.59
       ledger       17"        11"        43.18     27.94
       legal        8.5"       14"        21.59     35.56
       letter       8.5"       11"        21.59     27.94
       note         7.5"       10"        19.05     25.4
       
       Paper Keywords and paper size in points
       =======================================

       Letter           612x792
       LetterSmall      612x792
       Tabloid          792x1224
       Ledger          1224x792
       Legal            612x1008
       Statement        396x612
       Executive        540x720
       A0               2384x3371
       A1              1685x2384
       A2              1190x1684
       A3               842x1190
       A4               595x842
       A4Small          595x842
       A5               420x595
       B4               729x1032
       B5               516x729
       Envelope         ???x???
       Folio            612x936
       Quarto           610x780
       10x14            720x1008
*/


/*******************************************************************************/


/* Generated by reswrap from file landscape.gif */
const unsigned char landscape[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf1,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0xff,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,
  0x20,0x00,0x00,0x02,0x7c,0x84,0x8f,0xa9,0xcb,0xed,0x0f,0x63,0x0a,0xb4,0xda,0x8b,
  0xed,0x0a,0xa2,0xfb,0x0f,0x86,0x1d,0xa5,0x70,0xe2,0x29,0x72,0xc1,0x84,0xb6,0x9f,
  0xb9,0x1e,0xa6,0xdb,0xce,0x31,0x30,0xd3,0xc2,0x10,0x62,0x46,0xee,0x1a,0xf0,0x6a,
  0x3f,0xdd,0x4e,0xe8,0xb2,0xe8,0x84,0xc8,0x9a,0x80,0xb2,0x3c,0x26,0x9f,0x40,0xd1,
  0xb0,0x89,0x52,0x55,0x41,0x43,0xa9,0x13,0xda,0x62,0x32,0xa7,0xe0,0x53,0xd7,0x9b,
  0xa5,0xa2,0xce,0x47,0x36,0x48,0x6b,0xc6,0x8a,0xdd,0x1e,0xb8,0x51,0x67,0xbf,0x93,
  0xb7,0xfa,0xb7,0xba,0xff,0xc5,0x07,0x38,0x42,0x95,0x61,0x78,0x78,0xf1,0x54,0x24,
  0x01,0xa1,0x88,0x83,0x08,0x69,0xc8,0x38,0x49,0x59,0x69,0x79,0x89,0x99,0x09,0x51,
  0x00,0x00,0x3b
  };

/* Generated by reswrap from file portrait.gif */
const unsigned char portrait[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf1,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0xff,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,
  0x20,0x00,0x00,0x02,0x90,0x84,0x11,0xa9,0xcb,0x79,0x0f,0x4f,0x10,0xb4,0xda,0xeb,
  0xa2,0x9e,0xb7,0xd7,0x19,0x68,0x11,0xe7,0x75,0x5c,0x28,0x4a,0xa5,0x47,0xa2,0x29,
  0xb9,0x7e,0x4c,0x8a,0xc4,0xb1,0x22,0xc2,0x36,0x26,0xb8,0x90,0xbe,0xfb,0xf4,0x72,
  0xbb,0x01,0x6b,0xb8,0xb1,0x0d,0x8c,0x26,0xe4,0x28,0xb6,0x64,0xf2,0x7c,0x0f,0x60,
  0x25,0x2a,0xb5,0x80,0x88,0x25,0xe3,0xf2,0x48,0x55,0x79,0x98,0xdf,0x66,0xd8,0x60,
  0x95,0x96,0xa7,0xdc,0x0b,0x76,0xad,0x75,0xfe,0x3a,0x59,0xb8,0xf0,0x5c,0xb3,0x64,
  0x05,0x51,0x73,0x9b,0xbf,0xf6,0xc6,0x96,0x14,0xe4,0x47,0x58,0x18,0x87,0x67,0x15,
  0xb4,0x75,0x88,0x48,0xd1,0xf8,0xf4,0x78,0xf7,0xf7,0x18,0x39,0x37,0x09,0x29,0x57,
  0x95,0xa9,0xa9,0xd8,0xd9,0xb3,0x29,0xd1,0x40,0x5a,0x2a,0xca,0xd9,0x89,0x87,0x56,
  0xca,0xba,0xf0,0x50,0x00,0x00,0x3b
  };


/*******************************************************************************/

// Map
FXDEFMAP(FXPrintDialog) FXPrintDialogMap[]={
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_TO_PRINTER,FXPrintDialog::onCmdToPrinter),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_TO_PRINTER,FXPrintDialog::onUpdToPrinter),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_TO_FILE,FXPrintDialog::onCmdToFile),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_TO_FILE,FXPrintDialog::onUpdToFile),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_BROWSE_FILE,FXPrintDialog::onCmdBrowse),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_BROWSE_FILE,FXPrintDialog::onUpdBrowse),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_PROPERTIES,FXPrintDialog::onCmdProps),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_PROPERTIES,FXPrintDialog::onUpdProps),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_LANDSCAPE,FXPrintDialog::onCmdPortrait),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_LANDSCAPE,FXPrintDialog::onUpdPortrait),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_PORTRAIT,FXPrintDialog::onCmdLandscape),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_PORTRAIT,FXPrintDialog::onUpdLandscape),
  FXMAPFUNCS(SEL_COMMAND,FXPrintDialog::ID_PAGES_ALL,FXPrintDialog::ID_PAGES_RANGE,FXPrintDialog::onCmdPages),
  FXMAPFUNCS(SEL_UPDATE,FXPrintDialog::ID_PAGES_ALL,FXPrintDialog::ID_PAGES_RANGE,FXPrintDialog::onUpdPages),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_COLOR_PRINTER,FXPrintDialog::onCmdColor),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_COLOR_PRINTER,FXPrintDialog::onUpdColor),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_GRAY_PRINTER,FXPrintDialog::onCmdGray),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_GRAY_PRINTER,FXPrintDialog::onUpdGray),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_NUM_COPIES,FXPrintDialog::onCmdNumCopies),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_NUM_COPIES,FXPrintDialog::onUpdNumCopies),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_PAGES_FIRST,FXPrintDialog::onCmdFirstPage),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_PAGES_FIRST,FXPrintDialog::onUpdFirstPage),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_PAGES_LAST,FXPrintDialog::onCmdLastPage),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_PAGES_LAST,FXPrintDialog::onUpdLastPage),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_COLLATE_NORMAL,FXPrintDialog::onCmdCollateNormal),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_COLLATE_NORMAL,FXPrintDialog::onUpdCollateNormal),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_COLLATE_REVERSED,FXPrintDialog::onCmdCollateReversed),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_COLLATE_REVERSED,FXPrintDialog::onUpdCollateReversed),
  FXMAPFUNCS(SEL_COMMAND,FXPrintDialog::ID_SIZE_CUSTOM,FXPrintDialog::ID_SIZE_ENVELOPE,FXPrintDialog::onCmdMedia),
  FXMAPFUNCS(SEL_UPDATE,FXPrintDialog::ID_SIZE_CUSTOM,FXPrintDialog::ID_SIZE_ENVELOPE,FXPrintDialog::onUpdMedia),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_SIZE_WIDTH,FXPrintDialog::onCmdMediaWidth),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_SIZE_WIDTH,FXPrintDialog::onUpdMediaWidth),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_SIZE_HEIGHT,FXPrintDialog::onCmdMediaHeight),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_SIZE_HEIGHT,FXPrintDialog::onUpdMediaHeight),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_PRINTER_NAME,FXPrintDialog::onCmdPrinterName),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_PRINTER_NAME,FXPrintDialog::onUpdPrinterName),
  FXMAPFUNC(SEL_COMMAND,FXPrintDialog::ID_FILE_NAME,FXPrintDialog::onCmdFileName),
  FXMAPFUNC(SEL_UPDATE,FXPrintDialog::ID_FILE_NAME,FXPrintDialog::onUpdFileName),
  };

  
// Object implementation
FXIMPLEMENT(FXPrintDialog,FXDialogBox,FXPrintDialogMap,ARRAYNUMBER(FXPrintDialogMap))


// Separator item
FXPrintDialog::FXPrintDialog(FXWindow* owner,const FXString& name,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXDialogBox(owner,name,opts|DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE,x,y,w,h,4,4){
  FXVerticalFrame *contents=new FXVerticalFrame(this,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10, 10,10);
  
  // Close and cancel buttons
  FXHorizontalFrame* buttons=new FXHorizontalFrame(contents,LAYOUT_BOTTOM|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  new FXButton(buttons,"&Print",NULL,this,ID_ACCEPT,BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  new FXButton(buttons,"&Cancel",NULL,this,ID_CANCEL,FRAME_RAISED|FRAME_THICK|LAYOUT_TOP|LAYOUT_RIGHT,0,0,0,0,20,20);
  
  // Print destination
  FXGroupBox *dest=new FXGroupBox(contents,"Print Destination",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5, 10,5);
  
  sendtoprinter=new FXRadioButton(dest,"Pr&inter:",this,ID_TO_PRINTER,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  FXHorizontalFrame* printdest=new FXHorizontalFrame(dest,LAYOUT_SIDE_TOP|LAYOUT_FILL_X,0,0,0,0, 0,0,0,0, 10,10);
  printername=new FXComboBox(printdest,25,4,this,ID_PRINTER_NAME,COMBOBOX_STATIC|FRAME_SUNKEN|FRAME_THICK|LAYOUT_CENTER_Y|LAYOUT_FILL_X);
  new FXButton(printdest,"Properties...",NULL,this,ID_PROPERTIES,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_Y|LAYOUT_RIGHT|LAYOUT_FIX_WIDTH,0,0,100,0, 10,10);
  
  new FXFrame(dest,LAYOUT_SIDE_TOP|LAYOUT_FIX_HEIGHT,0,0,0,10);
  
  sendtofile=new FXRadioButton(dest,"&File:",this,ID_TO_FILE,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  FXHorizontalFrame* filedest=new FXHorizontalFrame(dest,LAYOUT_SIDE_TOP|LAYOUT_FILL_X,0,0,0,0, 0,0,0,0, 10,10);
  filename=new FXTextField(filedest,25,this,ID_FILE_NAME,FRAME_THICK|FRAME_SUNKEN|LAYOUT_CENTER_Y|LAYOUT_FILL_X);
  new FXButton(filedest,"&Browse...",NULL,this,ID_BROWSE_FILE,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_Y|LAYOUT_RIGHT|LAYOUT_FIX_WIDTH,0,0,100,0, 10,10);
  
  FXHorizontalFrame* bottom=new FXHorizontalFrame(contents,LAYOUT_BOTTOM|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,0,0,0,0);
  
  // Left
  FXVerticalFrame *left=new FXVerticalFrame(bottom,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,0,0,0,0, 10,10);
  
  // Pages
  FXGroupBox *pages=new FXGroupBox(left,"Pages",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  printall=new FXRadioButton(pages,"Print &All",this,ID_PAGES_ALL,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  printeven=new FXRadioButton(pages,"&Even Pages",this,ID_PAGES_EVEN,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  printodd=new FXRadioButton(pages,"&Odd Pages",this,ID_PAGES_ODD,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  printrange=new FXRadioButton(pages,"Print &Range:",this,ID_PAGES_RANGE,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  FXMatrix *pagerange=new FXMatrix(pages,2,MATRIX_BY_ROWS|LAYOUT_FILL_X|LAYOUT_SIDE_TOP,0,0,0,0, 20,0,0,0);
  new FXLabel(pagerange,"From:",NULL,LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_RIGHT);
  new FXLabel(pagerange,"To:",NULL,LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_RIGHT);
  firstpage=new FXSpinner(pagerange,4,this,ID_PAGES_FIRST,FRAME_THICK|FRAME_SUNKEN|LAYOUT_RIGHT);
  lastpage=new FXSpinner(pagerange,4,this,ID_PAGES_LAST,FRAME_THICK|FRAME_SUNKEN|LAYOUT_RIGHT);
  firstpage->setMinMax(1,10000000);
  lastpage->setMinMax(1,10000000);
  
  FXGroupBox *collate=new FXGroupBox(left,"Collate Order",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  firstpagefirst=new FXRadioButton(collate,"First Page First",this,ID_COLLATE_NORMAL,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  lastpagefirst=new FXRadioButton(collate,"Last Page First",this,ID_COLLATE_REVERSED,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  
  FXGroupBox *colors=new FXGroupBox(left,"Colors",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  printincolor=new FXRadioButton(colors,"Print in Color",this,ID_COLOR_PRINTER,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  printinblacknwhite=new FXRadioButton(colors,"Print in Black and White",this,ID_GRAY_PRINTER,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  
  // Right
  FXVerticalFrame *right=new FXVerticalFrame(bottom,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,0,0,0,0, 10,10);
  
  // Make some icons
  landscapeIcon=new FXGIFIcon(getApp(),landscape);
  portraitIcon=new FXGIFIcon(getApp(),portrait);
  
  // Layout
  FXGroupBox *copies=new FXGroupBox(right,"Copies",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  new FXLabel(copies,"Number of copies to print:",NULL,LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_RIGHT|LAYOUT_SIDE_LEFT);
  numberofcopies=new FXSpinner(copies,4,this,ID_NUM_COPIES,FRAME_THICK|FRAME_SUNKEN|LAYOUT_CENTER_Y|LAYOUT_SIDE_LEFT);
  numberofcopies->setMinMax(1,1000);
  
  // Layout
  FXGroupBox *orientation=new FXGroupBox(right,"Layout",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  new FXLabel(orientation,NULL,portraitIcon,LAYOUT_SIDE_LEFT|LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_CENTER_X|JUSTIFY_CENTER_Y|LAYOUT_CENTER_Y);
  orientportrait=new FXRadioButton(orientation,"Portrait",this,ID_PORTRAIT,LAYOUT_SIDE_LEFT|ICON_BEFORE_TEXT|JUSTIFY_CENTER_Y|LAYOUT_CENTER_Y);
  orientlanscape=new FXRadioButton(orientation,"Landscape",this,ID_LANDSCAPE,LAYOUT_SIDE_RIGHT|ICON_BEFORE_TEXT|JUSTIFY_CENTER_Y|LAYOUT_CENTER_Y);
  new FXLabel(orientation,NULL,landscapeIcon,LAYOUT_SIDE_RIGHT|LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_CENTER_X|JUSTIFY_CENTER_Y|LAYOUT_CENTER_Y);
  
  // Paper
  FXGroupBox *paperdims=new FXGroupBox(right,"Paper Size",GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_FILL_Y,0,0,0,0, 10,10,5,5);
  sizeletter=new FXRadioButton(paperdims,"Letter   (8.5 x 11 in)",this,ID_SIZE_LETTER,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  sizelegal=new FXRadioButton(paperdims,"Legal   (8.5 x 13 in)",this,ID_SIZE_LEGAL,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  sizeA4=new FXRadioButton(paperdims,"A4   (210 x 297 mm)",this,ID_SIZE_A4,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  sizeenvelope=new FXRadioButton(paperdims,"#10 Envelope   (4.13 x 9.5 in)",this,ID_SIZE_ENVELOPE,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  sizecustom=new FXRadioButton(paperdims,"Custom:",this,ID_SIZE_CUSTOM,LAYOUT_SIDE_TOP|ICON_BEFORE_TEXT);
  FXMatrix *papersize=new FXMatrix(paperdims,2,MATRIX_BY_ROWS|LAYOUT_FILL_X|LAYOUT_SIDE_TOP,0,0,0,0, 20,0,0,0);
  new FXLabel(papersize,"Width:",NULL,LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_RIGHT);
  new FXLabel(papersize,"Height:",NULL,LAYOUT_CENTER_Y|LAYOUT_RIGHT|JUSTIFY_RIGHT);
  sizewidth=new FXTextField(papersize,5,this,ID_SIZE_WIDTH,FRAME_THICK|FRAME_SUNKEN|LAYOUT_CENTER_Y|LAYOUT_FILL_X);
  sizeheight=new FXTextField(papersize,5,this,ID_SIZE_HEIGHT,FRAME_THICK|FRAME_SUNKEN|LAYOUT_CENTER_Y|LAYOUT_FILL_X);
  
  // Fill up with some defaults
  sendtoprinter->setCheck(TRUE);
  printall->setCheck(TRUE);
  firstpagefirst->setCheck(TRUE);
  printinblacknwhite->setCheck(TRUE);
  orientportrait->setCheck(TRUE);
  sizeletter->setCheck(TRUE);
  
  // Init
  printer.name[0]=0;
  printer.firstpage=1;
  printer.lastpage=1;
  printer.currentpage=1;
  printer.frompage=1;
  printer.topage=1;
  printer.mediasize=MEDIA_USLETTER;
  printer.mediawidth=612.0;
  printer.mediaheight=792.0;
  printer.leftmargin=18.0;
  printer.rightmargin=18.0;
  printer.topmargin=18.0;
  printer.bottommargin=18.0;
  printer.numcopies=1;
  printer.flags=0;
  }


// Build a list of printers
void FXPrintDialog::create(){
  FXDialogBox::create();
  
  char name[1000];

#ifndef WIN32
  
  FILE *pc;
  char buf[1000];
  int i;
  pc=fopen("/etc/printcap","r");
  if(!pc) return;
  name[0]=0;
  while(fgets(buf,1000,pc)){

    // Extra info if printcap has been generated by Printtool
    if(strncmp(buf,"##PRINTTOOL3##",14)==0){
      if(sscanf(buf,"%*s %*s %*s %*s %*s %*s %s",name)!=1) name[0]=0;
      continue;
      }
    
    // Blank line
    if(buf[0]=='#' || isspace(buf[0])){
      continue;
      }
    
    // Snarf printer name
    for(i=0; i<1000 && buf[i]!=0 && buf[i]!=':'; i++);
    buf[i]=0;
    
    // Append human-readable info, if any
    if(name[0]){
      strcat(buf," (");
      strcat(buf,name);
      strcat(buf,")");
      }
    
    // Add printer name
    if(strlen(buf)) printername->appendItem(buf);
    
    name[0]=0;
    }
  fclose(pc);
  
#else
  
  // What operating system are we running?
  OSVERSIONINFO osvi;
  osvi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
  GetVersionEx(&osvi);
  FXuint p;

#if !defined(__CYGWIN__) && !defined(__MINGW32__) /* EnumPrinters is not in their headers :( */

  if(osvi.dwPlatformId==VER_PLATFORM_WIN32_NT){           // WinNT
    DWORD dwFlags=PRINTER_ENUM_LOCAL|PRINTER_ENUM_CONNECTIONS;
    DWORD dwBytesNeeded,dwNumPrinters;
    EnumPrinters(dwFlags,NULL,4,NULL,0,&dwBytesNeeded,&dwNumPrinters);
    PRINTER_INFO_4 *prtinfo;
    if(FXMALLOC(&prtinfo,BYTE,dwBytesNeeded)){
      if(EnumPrinters(dwFlags,NULL,4,(LPBYTE)prtinfo,dwBytesNeeded,&dwBytesNeeded,&dwNumPrinters)){
	for(p=0; p<dwNumPrinters; p++) printername->appendItem(prtinfo[p].pPrinterName);
	}
      FXFREE(&prtinfo);
      }
    }
  else if(osvi.dwPlatformId==VER_PLATFORM_WIN32_WINDOWS){ // Win95 or Win98
    FXASSERT(FALSE);                                      // This hasn't been tested on Win95/98!
    DWORD dwFlags=PRINTER_ENUM_LOCAL;
    DWORD dwBytesNeeded,dwNumPrinters;
    EnumPrinters(dwFlags,NULL,5,NULL,0,&dwBytesNeeded,&dwNumPrinters);
    PRINTER_INFO_5 *prtinfo;
    if(FXMALLOC(&prtinfo,BYTE,dwBytesNeeded)){
      if(EnumPrinters(dwFlags,NULL,5,(LPBYTE)prtinfo,dwBytesNeeded,&dwBytesNeeded,&dwNumPrinters)){
        for(p=0; p<dwNumPrinters; p++) printername->appendItem(prtinfo[p].pPrinterName);
	}
      FXFREE(&prtinfo);
      }
    }
  else
    FXASSERT(FALSE);

#endif
  
#endif
  
  // Fill in default printer name
  if(0<printername->getNumItems()){
    printername->setCurrentItem(0);
    }
  
  // Fill in default filename
  fxabspath(name,NULL,"postscript.ps");
  const FXchar *printfile=getApp()->reg().readStringEntry("PRINTER","printtofile",name);
  filename->setText(printfile);
  
  // Force it
  handle(this,MKUINT(ID_TO_PRINTER,SEL_COMMAND),NULL);
  }


// Send to printer 
long FXPrintDialog::onCmdToPrinter(FXObject*,FXSelector,void*){
  FXString name=printername->getText();
  if(!name.empty()){
    FXint pos=name.findf(' ');
    if(pos>=0){ name.remove(pos,1000); }
    strncpy(printer.name,name.text(),sizeof(printer.name));
    FXTRACE((100,"Print to printer: %s\n",printer.name));
    }
  printer.flags&=~PRINT_DEST_FILE;
  return 1;
  }


// Update send to printer 
long FXPrintDialog::onUpdToPrinter(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_DEST_FILE)
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Send to file 
long FXPrintDialog::onCmdToFile(FXObject*,FXSelector,void*){
  FXString name=filename->getText();
  if(!name.empty()){
    strncpy(printer.name,name.text(),sizeof(printer.name));
    FXTRACE((100,"Print to file: %s\n",printer.name));
    }
  printer.flags|=PRINT_DEST_FILE;
  return 1;
  }


// Update send to file 
long FXPrintDialog::onUpdToFile(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_DEST_FILE)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Browse output file 
long FXPrintDialog::onCmdBrowse(FXObject*,FXSelector,void*){
  FXFileDialog filedialog(this,"Select Output File");
  filedialog.setFilename(filename->getText());
  if(filedialog.execute()){
    if(fxexists(filedialog.getFilename().text())){
      if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite file?","Overwrite existing output file?")){
        return 1;
        }
      }
    FXString name=filedialog.getFilename();
    if(!name.empty()){
      strncpy(printer.name,name.text(),sizeof(printer.name));
      filename->setText(name);
      getApp()->reg().writeStringEntry("PRINTER","printtofile",printer.name);
      FXTRACE((100,"Print to file: %s\n",printer.name));
      }
    }
  return 1;
  }


// Update browse output file 
long FXPrintDialog::onUpdBrowse(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_DEST_FILE)
    sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL);
  return 1;
  }


// Printer properties panel 
long FXPrintDialog::onCmdProps(FXObject*,FXSelector,void*){
  return 1;
  }


// Update printer properties panel
long FXPrintDialog::onUpdProps(FXObject* sender,FXSelector,void*){
  //if(printer.flags&PRINT_DEST_FILE)
    sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL);
  //else
  //  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  return 1;
  }


// File name 
long FXPrintDialog::onCmdFileName(FXObject*,FXSelector,void*){
  if(printer.flags&PRINT_DEST_FILE){
    FXString name=filename->getText();
    if(!name.empty()){
      strncpy(printer.name,name.text(),sizeof(printer.name));
      getApp()->reg().writeStringEntry("PRINTER","printtofile",printer.name);
      FXTRACE((100,"Print to file: %s\n",printer.name));
      }
    }
  return 1;
  }


// Update File name 
long FXPrintDialog::onUpdFileName(FXObject* sender,FXSelector,void* ptr){
//   if(printer.flags&PRINT_DEST_FILE){
//     sender->handle(this,MKUINT(ID_SETVALUE,SEL_COMMAND),(void*)printer.name);
//     }
  return 1;
  }


// Printer name 
long FXPrintDialog::onCmdPrinterName(FXObject*,FXSelector,void*){
  if(!(printer.flags&PRINT_DEST_FILE)){
    FXString name=printername->getText();
    if(!name.empty()){
      //FXint pos=name.findf(' ');
      //if(pos>=0){ name.remove(pos,1000); }
      name=name.before(' ',1);
      strncpy(printer.name,name.text(),sizeof(printer.name));
      FXTRACE((100,"Print to printer: %s\n",printer.name));
      }
    }
  return 1;
  }


// Update Printer name 
long FXPrintDialog::onUpdPrinterName(FXObject* sender,FXSelector,void* ptr){
//   if(!(printer.flags&PRINT_DEST_FILE)){
//     sender->handle(this,MKUINT(ID_SETVALUE,SEL_COMMAND),(void*)printer.name);
//     }
  return 1;
  }


// Portrait 
long FXPrintDialog::onCmdPortrait(FXObject*,FXSelector,void* ptr){
  printer.flags&=~PRINT_LANDSCAPE;
  return 1;
  }


// Update portrait
long FXPrintDialog::onUpdPortrait(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_LANDSCAPE)
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Landscape 
long FXPrintDialog::onCmdLandscape(FXObject*,FXSelector,void* ptr){
  printer.flags|=PRINT_LANDSCAPE;
  return 1;
  }


// Update landscape
long FXPrintDialog::onUpdLandscape(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_LANDSCAPE)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Pages 
long FXPrintDialog::onCmdPages(FXObject*,FXSelector sel,void* ptr){
  FXuint sid=SELID(sel);
  switch(sid){
    case ID_PAGES_ALL: 
      printer.flags&=~(PRINT_PAGES_EVEN|PRINT_PAGES_ODD|PRINT_PAGES_RANGE);
      printer.frompage=printer.firstpage;
      printer.topage=printer.lastpage;
      break;
    case ID_PAGES_EVEN: 
      printer.flags|=PRINT_PAGES_EVEN; 
      printer.flags&=~(PRINT_PAGES_ODD|PRINT_PAGES_RANGE);
      printer.frompage=(printer.firstpage+1)&~1;    // Get even 
      printer.topage=printer.lastpage&~1;
      break;
    case ID_PAGES_ODD: 
      printer.flags|=PRINT_PAGES_ODD;
      printer.flags&=~(PRINT_PAGES_EVEN|PRINT_PAGES_RANGE);
      printer.frompage=printer.firstpage|1;         // Get odd?
      printer.topage=(printer.lastpage-1)|1;
      break;
    case ID_PAGES_RANGE: 
      printer.flags|=PRINT_PAGES_RANGE; 
      printer.flags&=~(PRINT_PAGES_EVEN|PRINT_PAGES_ODD);
      printer.frompage=printer.firstpage;
      printer.topage=printer.lastpage;
      break;
    }
  if(printer.frompage>printer.lastpage) printer.frompage=printer.lastpage;
  if(printer.frompage<printer.firstpage) printer.frompage=printer.firstpage;
  if(printer.topage>printer.lastpage) printer.topage=printer.lastpage;
  if(printer.topage<printer.firstpage) printer.topage=printer.firstpage;
  return 1;
  }


// Update Pages
long FXPrintDialog::onUpdPages(FXObject* sender,FXSelector sel,void*){
  FXuint ch=ID_UNCHECK;
  FXuint sid=SELID(sel);
  switch(sid){
    case ID_PAGES_ALL: 
      if(!(printer.flags&(PRINT_PAGES_EVEN|PRINT_PAGES_ODD|PRINT_PAGES_RANGE))) ch=ID_CHECK;
      break;
    case ID_PAGES_EVEN: 
      if(printer.flags&PRINT_PAGES_EVEN) ch=ID_CHECK;
      break;
    case ID_PAGES_ODD: 
      if(printer.flags&PRINT_PAGES_ODD) ch=ID_CHECK;
      break;
    case ID_PAGES_RANGE: 
      if(printer.flags&PRINT_PAGES_RANGE) ch=ID_CHECK;
      break;
    }
  sender->handle(this,MKUINT(ch,SEL_COMMAND),NULL);
  return 1;
  }


// Color 
long FXPrintDialog::onCmdColor(FXObject*,FXSelector,void* ptr){
  printer.flags|=PRINT_COLOR;
  return 1;
  }


// Update color
long FXPrintDialog::onUpdColor(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_COLOR)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Gray 
long FXPrintDialog::onCmdGray(FXObject*,FXSelector,void* ptr){
  printer.flags&=~PRINT_COLOR;
  return 1;
  }


// Update gray
long FXPrintDialog::onUpdGray(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_COLOR)
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Number of copies 
long FXPrintDialog::onCmdNumCopies(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_GETINTVALUE,SEL_COMMAND),(void*)&printer.numcopies);
  return 1;
  }


// Update number of copies
long FXPrintDialog::onUpdNumCopies(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_SETINTVALUE,SEL_COMMAND),(void*)&printer.numcopies);
  return 1;
  }


// First page 
long FXPrintDialog::onCmdFirstPage(FXObject* sender,FXSelector,void*){
  FXint pg;
  sender->handle(this,MKUINT(ID_GETINTVALUE,SEL_COMMAND),(void*)&pg);
  FXASSERT(pg>0);
  if((FXuint)pg<printer.firstpage) pg=printer.firstpage;
  if((FXuint)pg>printer.lastpage) pg=printer.lastpage;
  printer.frompage=pg;
  return 1;
  }


// Update first page
long FXPrintDialog::onUpdFirstPage(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_SETINTVALUE,SEL_COMMAND),(void*)&printer.frompage);
  return 1;
  }


// Last page 
long FXPrintDialog::onCmdLastPage(FXObject* sender,FXSelector,void*){
  FXint pg;
  sender->handle(this,MKUINT(ID_GETINTVALUE,SEL_COMMAND),(void*)&pg);
  FXASSERT(pg>0);
  if((FXuint)pg<printer.firstpage) pg=printer.firstpage;
  if((FXuint)pg>printer.lastpage) pg=printer.lastpage;
  printer.topage=pg;
  return 1;
  }


// Update last page
long FXPrintDialog::onUpdLastPage(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_SETINTVALUE,SEL_COMMAND),(void*)&printer.topage);
  return 1;
  }


// Collate Normal 
long FXPrintDialog::onCmdCollateNormal(FXObject* sender,FXSelector,void*){
  printer.flags&=~PRINT_COLLATE_REVERSED;
  return 1;
  }


// Update Collate Normal 
long FXPrintDialog::onUpdCollateNormal(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_COLLATE_REVERSED)
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Collate Reversed 
long FXPrintDialog::onCmdCollateReversed(FXObject* sender,FXSelector,void*){
  printer.flags|=PRINT_COLLATE_REVERSED;
  return 1;
  }


// Update Collate Reversed 
long FXPrintDialog::onUpdCollateReversed(FXObject* sender,FXSelector,void*){
  if(printer.flags&PRINT_COLLATE_REVERSED)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Paper width 
long FXPrintDialog::onCmdMediaWidth(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_GETREALVALUE,SEL_COMMAND),(void*)&printer.mediawidth);
  return 1;
  }


// Update Paper width
long FXPrintDialog::onUpdMediaWidth(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_SETREALVALUE,SEL_COMMAND),(void*)&printer.mediawidth);
  return 1;
  }


// Paper height 
long FXPrintDialog::onCmdMediaHeight(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_GETREALVALUE,SEL_COMMAND),(void*)&printer.mediaheight);
  return 1;
  }


// Update Paper height
long FXPrintDialog::onUpdMediaHeight(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_SETREALVALUE,SEL_COMMAND),(void*)&printer.mediaheight);
  return 1;
  }


// Standard paper dimensions 
long FXPrintDialog::onCmdMedia(FXObject* sender,FXSelector sel,void*){
  FXuint sid=SELID(sel);
  switch(sid){
    case ID_SIZE_CUSTOM: 
      printer.mediasize=MEDIA_CUSTOM;
      break;
    case ID_SIZE_LETTER: 
      printer.mediasize=MEDIA_USLETTER;
      printer.mediawidth=612.0;
      printer.mediaheight=792.0;
      printer.leftmargin=18.0;
      printer.rightmargin=18.0;
      printer.topmargin=18.0;
      printer.bottommargin=18.0;
      break;
    case ID_SIZE_LEGAL: 
      printer.mediasize=MEDIA_LEGAL;
      printer.mediawidth=612.0;
      printer.mediaheight=1008.0;
      printer.leftmargin=11.0;
      printer.rightmargin=11.0;
      printer.topmargin=15.0;
      printer.bottommargin=15.0;
      break;
    case ID_SIZE_A4: 
      printer.mediasize=MEDIA_A4;
      printer.mediawidth=595.0;
      printer.mediaheight=842.0;
      printer.leftmargin=14.0;
      printer.rightmargin=14.0;
      printer.topmargin=14.0;
      printer.bottommargin=14.0;
      break;
    case ID_SIZE_ENVELOPE: 
      printer.mediasize=MEDIA_ENVELOPE;
      printer.mediawidth=297.0;
      printer.mediaheight=684.0;
      printer.leftmargin=10.0;
      printer.rightmargin=10.0;
      printer.topmargin=12.0;
      printer.bottommargin=12.0;
      break;
    }
  return 1;
  }


// Update Standard paper dimensions
long FXPrintDialog::onUpdMedia(FXObject* sender,FXSelector sel,void*){
  FXuint ch=ID_UNCHECK;
  FXuint sid=SELID(sel);
  switch(sid){
    case ID_SIZE_CUSTOM: 
      if(printer.mediasize==MEDIA_CUSTOM) ch=ID_CHECK;
      break;
    case ID_SIZE_LETTER: 
      if(printer.mediasize==MEDIA_USLETTER) ch=ID_CHECK;
      break;
    case ID_SIZE_LEGAL: 
      if(printer.mediasize==MEDIA_LEGAL) ch=ID_CHECK;
      break;
    case ID_SIZE_A4: 
      if(printer.mediasize==MEDIA_A4) ch=ID_CHECK;
      break;
    case ID_SIZE_ENVELOPE: 
      if(printer.mediasize==MEDIA_ENVELOPE) ch=ID_CHECK;
      break;
    }
  sender->handle(this,MKUINT(ch,SEL_COMMAND),NULL);
  return 1;
  }


// Set printer to print to
void FXPrintDialog::setPrinter(const FXPrinter& pr){
  printer=pr;
  }


// Obtain printer to print to
void FXPrintDialog::getPrinter(FXPrinter& pr){
  pr=printer;
  }


// Save data
void FXPrintDialog::save(FXStream& store) const {
  FXDialogBox::save(store);
  }


// Load data
void FXPrintDialog::load(FXStream& store){ 
  FXDialogBox::load(store);
  }


// Cleanup
FXPrintDialog::~FXPrintDialog(){
  delete landscapeIcon;
  delete portraitIcon;
  landscapeIcon=(FXGIFIcon*)-1;
  portraitIcon=(FXGIFIcon*)-1;
  }


